# -*- coding: utf-8 -*-
import logging
import requests
from datetime import datetime

from odoo import modules
from odoo import api, fields, models
from odoo.addons.saas_worker.auth_util import SaasAuth
from odoo.exceptions import AccessError, AccessDenied, ValidationError
from odoo.release import version as server_version
from odoo.tools.parse_version import parse_version
from odoo.tools.translate import _
from odoo.tools import convert_file
from odoo.tools import config
from odoo.addons.saas_worker import metabase
from odoo.addons.saas_worker.const import EMAIL_DAILY_LIMIT_MASS_MAIL, SPECIAL_PREFIXES, WEBSITE, TIMEOUT
from odoo.addons.saas_worker.util import version_gte, version_lte

_logger = logging.getLogger(__name__)

def version_lt(than):
    return parse_version(server_version) < parse_version(str(than))

ALLOWED_THEMES = [
    "theme_bistro",
    "theme_bootswatch",
    "theme_clean",
    "theme_common",
    "theme_default",
    "theme_enark",
    "theme_graphene",
    "theme_kiddo",
    "theme_kiddo_sale",
    "theme_loftspace",
    "theme_loftspace_sale",
    "theme_nano",
    "theme_treehouse",
    "theme_zap",
]

RESTRICTED_MODULES = [
    'saas_base_import',
]

if version_lte('saas~16.3'):
    RESTRICTED_MODULES += [
        'saas_whatsapp',
    ]
if version_lte("saas~17.1"):
    RESTRICTED_MODULES += [
        'saas_product_barcodelookup'
    ]


class IrModuleModule(models.Model):
    _inherit = 'ir.module.module'

    if version_gte("17.0"):
        is_module_official = fields.Boolean(compute='_compute_is_module_official')

        def _compute_is_module_official(self):
            # if `base_import_module` is installed, determine if module is official
            # using the `module_type` field, otherwise consider them official by
            # default on saas.
            self.is_module_official = True
            if 'module_type' in self._fields:
                for module in self:
                    module.is_module_official = module.module_type == 'official'

        def _register_hook(self):
            super()._register_hook()

            if not (config['init'] or config['update']) and not self.env.registry.ready:
                # on normal startup, ensure apps view are up-to-date
                # (3 queries for check if up-to-date [IMD + exists() + view read])
                st_app_view_kanban = self.env.ref('saas_trial.paid_apps_module_kanban',
                                                  raise_if_not_found=False)
                if (
                    st_app_view_kanban and
                    'is_module_official' not in st_app_view_kanban.arch_base
                ):
                    try:
                        # try locking view row or skip update if already locked by another worker
                        self.env.cr.execute(
                            "SELECT id FROM ir_ui_view WHERE id = %s FOR UPDATE SKIP LOCKED",
                            (st_app_view_kanban.id,)
                        )
                        if self.env.cr.rowcount:
                            st_app_view_kanban.reset_arch(mode='hard')  # reload from fs
                            _logger.info('auto-resetted view %s (%s) for OPW-3698982',
                                         st_app_view_kanban.id, st_app_view_kanban.name)
                    except ValidationError:
                        _logger.exception('could not auto-reset view %s (%s)',
                                          st_app_view_kanban.id, st_app_view_kanban.name)

    # Prevent external call with forged context
    __courtois = object()

    def _createbase_automation_flag(self):
        if not version_gte("17.0"):
            return

        self.env['ir.model.data'].create({
            'module': 'base_automation',
            'name': 'installed-17',
            'model': 'ir.ui.view',
            'res_id': 0,
            'noupdate': True,
        })

    @api.multi
    def _button_immediate_function(self, function):
        ba_module = self.env.ref('base.module_base_automation')
        ba_was_installed = ba_module.state == 'installed'

        res = super(IrModuleModule, self)._button_immediate_function(function)

        ba_is_installed = ba_module.state == 'installed'
        if not ba_was_installed and ba_is_installed:
            self._createbase_automation_flag()

        with self.env.cr.savepoint() as sp:  # ignore possible failed cursor if concurrent update (e.g. on ir.config_parameter)
            if not self.env['publisher_warranty.contract'].sudo().with_context(force_ping=True).update_notification(cron_mode=True):
                sp.rollback()
        return res

    @api.multi
    def button_immediate_install(self):

        # Allow 1 paying dep if only website is installed and module is a brige app with website
        allowed_apps = self.env['ir.module.module'].search([
            ('application', '=', True),
            ('name', 'ilike', 'website_')
        ]).mapped('name')

        # Installable via v15 website configurator, but not an app
        allowed_apps.append('website_event_sale')

        # Check if we need to keep the "1-App Free" state
        if len(self) == 1 and self.name in allowed_apps:

            installed_apps = self.search([('state', '=', 'installed')])
            paid_installed_apps = installed_apps.filtered(lambda m: m.application)

            # mail is an app, but it's a dependency of website
            if all(pia.name in ['website', 'mail'] for pia in paid_installed_apps):
                # Keeper of the "1-App Free"
                self = self.with_context({'ignore_extra_apps': self.__courtois})

        return super(IrModuleModule, self).button_immediate_install()

    # overridden to catch all attempts to uninstall this module
    @api.multi
    def write(self, values):
        mods = self
        mod_names = [mod.name for mod in mods]
        new_state = values.get('state')
        if new_state == 'to remove' and 'saas_trial' in mod_names:
            raise AccessError(_("Module 'saas_trial' cannot be uninstalled"))
        elif new_state == 'to install' and any(name.startswith('hw_') for name in mod_names):
            raise AccessError(_('PosBox hardware proxy modules cannot be installed on Odoo Online. '
                                'They should be installed on the PosBox or the computer with the '
                                'devices'))
        elif new_state == 'to install':

            paying_deps = self.get_paying_deps(module_ids=mods.ids)
            db = metabase.get(self._cr.dbname)
            if paying_deps and any([x.name in paying_deps for x in mods]):
                # Check if we can bypass the extra_apps check
                if db and db.mode == 'trial' and not db.extra_apps \
                        and self._context.get('ignore_extra_apps') is not self.__courtois:
                    # out of the Free For One App World
                    # only set expire once
                    metabase.update(db, extra_apps=True)
                    if db.status != 'to_activate' and not db.name.startswith(SPECIAL_PREFIXES):
                        metabase.set_expiry_in(db.name, '15 days')

            # Reduce the email daily limit for db recently created
            if (
                db
                and db.mode == 'trial'
                and not db.name.startswith(SPECIAL_PREFIXES)
                and 'mass_mailing' in mod_names
                and (datetime.now() - db.create_date).total_seconds() < 48 * 3600
            ):
                metabase.update(db, email_daily_limit=EMAIL_DAILY_LIMIT_MASS_MAIL)

            # Disable HSTS when point_of_sale is installed because epos printer
            # and six terminals need http with the default setup
            if (
                db
                and db.mode in ['trial', 'paid']
                and 'point_of_sale' in mod_names
            ):
                metabase.update(db, secure=False)

        return super(IrModuleModule, self).write(values)

    def _get_pricing(self):
        db = metabase.get(self._cr.dbname)
        # at the database installation, its entry in meta is not created yet
        if not db:
            return False

        pricing = False
        if db.mode == "trial":
            pricing = "1appfree" if not db.extra_apps else "trial"
        else:
            # paid: we need to call accounts to get subscribed plan
            try:
                params = {'params': {'db_uuid': db.uuid}}
                r = requests.post(WEBSITE + '/odoo-enterprise/get-pricing', json=params, auth=SaasAuth(), timeout=TIMEOUT)
                pricing = r.json().get('result', False)
            except Exception:
                _logger.warning("Could not retrieve paid apps from accounts", exc_info=True)
                pricing = False
        return pricing

    # override needed to remove RESTRICTED_MODULES, otherwise update_list()
    # will crash while trying to locate the module
    @classmethod
    def get_module_info(cls, name):
        if name in RESTRICTED_MODULES:
            return {}
        return super().get_module_info(name)

    @api.model
    def install_from_urls(self, urls):
        raise AccessDenied()

    if version_gte("saas~16.2"):  # no more count argument
        @api.model  # web_search_read use _search and not search
        def _search(self, args, offset=0, limit=None, order=None, access_rights_uid=None):
            if RESTRICTED_MODULES:
                args += ['|', ('name', 'not in', RESTRICTED_MODULES),
                        ('state', 'in', ['installed', 'to upgrade', 'to install', 'to remove'])]
            return super(IrModuleModule, self)._search(
                args, offset=offset, limit=limit, order=order, access_rights_uid=access_rights_uid
            )
    else:
        @api.model
        def search(self, args, offset=0, limit=0, order=None, count=False):
            if RESTRICTED_MODULES:
                args += ['|', ('name', 'not in', RESTRICTED_MODULES),
                        ('state', 'in', ['installed', 'to upgrade', 'to install', 'to remove'])]
            return super(IrModuleModule, self).search(args, offset=offset, limit=limit, order=order, count=count)

    @api.model
    def create(self, values):
        # As `search` ignore some themes, `update_list` think some module does not exists and
        # try to recreate them. Find the existing module id.
        name = values.get('name')
        if name and (name.startswith('theme_') or (name in RESTRICTED_MODULES)):
            mod_path = modules.get_module_path(name)
            if (mod_path and '/themes/' in mod_path) or (name in RESTRICTED_MODULES):
                self._cr.execute("SELECT id FROM ir_module_module WHERE name=%s", [name])
                if self._cr.rowcount:
                    return self.browse(self._cr.fetchone()[0])
        return super(IrModuleModule, self).create(values)

    @api.model
    def get_paying_deps(self, module_ids=None, module_names=None, paid_feature=None, **kwargs):
        if paid_feature:
            current_pricing = self._get_pricing()
            included_in_current_plan = False
            if paid_feature in ("res.company", "ir.model.fields", "ir.actions.server"):
                # only warn if we need to upgrade subscription to "custom" plan
                included_in_current_plan = current_pricing not in ("trial", "standard")
            elif paid_feature == "res.users":
                # always warn for user when out of 1-app-free mode
                included_in_current_plan = current_pricing == "1appfree"

            # Note: as paid features are not related to a specific module, if
            #   it's not covered by the current subscription plan we return a
            #   "non-empty" module list (here we arbitrarily chose "base") so
            #   that the confirmation modal is displayed to the user.
            return ["base"] if not included_in_current_plan else []

        if not module_names and not module_ids:
            return []
        if module_ids:
            modules = self.browse(module_ids)
        else:
            modules = self.search([('name', 'in', module_names or [])])

        # filter the already installed modules
        modules = modules.filtered(lambda x: x.state == 'uninstalled')
        if not modules:
            return []

        deps = modules.upstream_dependencies()
        # we now search for auto_install modules that could be installed
        auto_install_modules = self.search([('state', '=', 'uninstalled'), ('auto_install', '=', True)])

        new_modules = True
        while new_modules:
            # we want to avoid infinite loop, so we will set to True only if we are sure
            new_modules = False

            for m in auto_install_modules:
                visited_modules = deps | modules
                if m not in visited_modules:
                    # retrieve the direct (1st layer) non installed dependencies of the auto install module
                    module_first_dependencies = m.mapped('dependencies_id.depend_id').filtered(lambda x: x.state == 'uninstalled')

                    if (module_first_dependencies & visited_modules) == module_first_dependencies:
                        new_modules = True
                        module_all_dependencies = m.upstream_dependencies()
                        deps |= module_all_dependencies | m

        # Fetch pricing linked to db to determine what modules are paid
        current_pricing = self._get_pricing()
        mods_to_install = deps | modules
        # Standard Pricing, only web_studio is a paying dep
        if current_pricing == 'standard':
            if 'web_studio' in mods_to_install.mapped('name'):
                return ['web_studio']
            # base_automation becomes a paying dep in v17
            elif version_gte("17.0") and 'base_automation' in mods_to_install.mapped('name'):
                return ['base_automation']
            else:
                return []
        # Already paying the maximum
        elif current_pricing == 'custom':
            return []
        # 1-App Free, any module that has `application` set to True is a paying dep
        else:
            return mods_to_install.filtered(lambda m: m.application).mapped('name')

    @api.model
    def _load_xml(self, module, filename, min_version=False, max_version=False, mode='update'):
        """
            Allow to load xml file only if the min-version <= current version <= max_version
            Designed to be used in tag function in xml file

            :param mode: Load the file in the given mode (update or init). Default is update mode.

            If you only want to load data at module install
            insert function tag in no update section
        """
        if not version_gte(min_version) or (max_version and not version_lte(max_version)):
            return

        _logger.info("loading %s/%s", module, filename)
        if version_gte("saas~16.2"):
            convert_file(self.env, module, filename, {}, mode=mode)
        else:
            convert_file(self.env.cr, module, filename, {}, mode=mode)


class IrModuleModuleDependency(models.Model):
    _inherit = 'ir.module.module.dependency'

    if version_gte("saas~16.2"):  # no more count argument
        @api.model  # web_search_read use _search and not search
        def _search(self, args, offset=0, limit=None, order=None, access_rights_uid=None):
            if RESTRICTED_MODULES:
                args += [('module_id.name', 'not in', RESTRICTED_MODULES)]
            return super()._search(
                args, offset=offset, limit=limit, order=order, access_rights_uid=access_rights_uid
            )
    else:
        @api.model
        def search(self, args, offset=0, limit=0, order=None, count=False):
            if RESTRICTED_MODULES:
                args += [('module_id.name', 'not in', RESTRICTED_MODULES)]
            return super().search(args, offset=offset, limit=limit, order=order, count=count)
